@@ -1,27 +1,11 @@
 from collections import OrderedDict
+from contextlib import contextmanager
 from datetime import datetime, timedelta
+from sys import getsizeof
+from threading import Lock
 from typing import Any, Dict, Optional
 
-
-class CacheValue:
-    """Dummy class for handling cache values."""
-
-    def __init__(self, value: Any, expire_at: datetime = None):
-        self.value = value
-        self.expire_at = expire_at
-        self.access = 0
-
-    def expired(self):
-        """Check if value is expired."""
-        if self.expire_at:
-            return datetime.utcnow() > self.expire_at
-        return False  # pragma: no cover
-
-    def refresh_ttl(self, expire_at: datetime):
-        self.expire_at = expire_at
-
-    def __eq__(self, otherinstance: "CacheValue"):
-        return self.value == otherinstance.value
+from onecache.cache_value import CacheValue
 
 
 class ExpirableCache(object):
@@ -34,34 +18,68 @@ class ExpirableCache(object):
         * **timeout (int)**: Timeout in milliseconds, if it is None,
             there is no timeout. default=None
         * **refresh_ttl (int)**: Refresh ttl anytime key is accessed. default=False
+        * **thread_safe (bool)**: Tell cache decorator to be thread safe. default=False
+        * **max_mem_size (int)**: max mem size inside cache structure. default=None which means no limit
     """
 
-    def __init__(self, size=512, timeout=None, refresh_ttl=False):
+    def __init__(
+        self,
+        size: int = 512,
+        timeout: int = None,
+        refresh_ttl=False,
+        thread_safe=False,
+        max_mem_size: int = None,
+    ):
         self.cache: Dict[str, CacheValue] = OrderedDict()
         self.timeout = timeout
         self.size = size
+        self.max_mem_size = max_mem_size
+        self.current_size = 0
         self.refresh_ttl = refresh_ttl
+        self.lock = None
+        if thread_safe:
+            self.lock = Lock()
 
     def set(self, key, data):
-        if len(self.cache) + 1 > self.size and key not in self.cache:
-            self._pop_one()
+        with self._scoped():
+            if len(self.cache) + 1 > self.size and key not in self.cache:
+                self._pop_one()
+            value = None
 
-        if self.timeout:
-            expire_at = datetime.utcnow() + timedelta(milliseconds=self.timeout)
-            self.cache[key] = CacheValue(data, expire_at)
-        else:
-            self.cache[key] = CacheValue(data)
+            if self.timeout:
+                expire_at = datetime.utcnow() + timedelta(milliseconds=self.timeout)
+                value = CacheValue(data, expire_at)
+            else:
+                value = CacheValue(data)
+
+            if self.max_mem_size:
+                # pop data if it will exceed max mem size
+                while (
+                    self.current_size + value.size > self.max_mem_size
+                    and len(self.cache) > 1
+                ):
+                    self._pop_one()
+            self.cache[key] = value
+
+            if self.max_mem_size:
+                self.current_size += value.size
 
     def _pop_one(self):
-        self.cache.pop(next(iter(self.cache)))
+        key = next(iter(self.cache))
+        val = self.cache.pop(key)
+        if self.max_mem_size:
+            self.current_size -= val.size
 
-    def get(self, key):
+    def get(self, key, fetch_value=True):
         self._check_expired(key)
         cache_value = self.cache.get(key)
         if cache_value:
             if self.refresh_ttl:
-                self.refresh_key_ttl(key)
-            return cache_value.value
+                with self._scoped():
+                    self.refresh_key_ttl(key)
+            if fetch_value:
+                return cache_value.value
+            return cache_value
 
     def _check_expired(self, key):
         to_rm = []
@@ -73,6 +91,9 @@ class ExpirableCache(object):
             self._remove_key(key)
 
     def _remove_key(self, key):
+        item = self.cache[key]
+        if self.max_mem_size:
+            self.current_size -= item.size
         del self.cache[key]
 
     def __contains__(self, key):
@@ -114,23 +135,30 @@ class ExpirableCache(object):
         if key in self.cache:
             self._remove_key(key)
 
+    @contextmanager
+    def _scoped(self):
+        if self.lock:
+            with self.lock:
+                yield
+        else:
+            yield
+
 
 class LRUCache(ExpirableCache):
     def get(self, key):
-        res = super().get(key)
-        if res:
-            cache_value = self.cache[key]
-            if cache_value:
+        cache_value = super().get(key, fetch_value=False)
+        if cache_value:
+            with self._scoped():
                 self._increment(cache_value)
-                return res
-        return res
+        return cache_value.value
 
     def _increment(self, val: CacheValue):
-        val.access += 1
+        val.increment()
 
     def _pop_one(self):
         ordered_cache = sorted(self.cache.items(), key=lambda item: item[1].access)
-        del self.cache[ordered_cache[0][0]]
+        key, cache_value = ordered_cache[0]
+        self._remove_key(key)
 
     @property
     def access(self):
@@ -149,17 +177,27 @@ class CacheDecorator:
         skip_args: bool = False,
         cache_class=LRUCache,
         refresh_ttl: Optional[bool] = False,
+        thread_safe: Optional[bool] = False,
+        max_mem_size: Optional[int] = None,
     ):
         """
         Args:
-            * **maxsize (int)**: Maximun size of cache. default: 512
+            * **maxsize (int)**: Maximun size of cache items. default: 512
             * **ttl (int)**: time to expire in milliseconds, if None, it does not expire. default: None
             * **skip_args (bool)**: apply cache as the function doesn't have any arguments, default: False
             * **cache_class (class)**: Class to use for cache instance. default: LRUCache
             * **refresh_ttl (bool)**: if cache with ttl, This flag makes key expiration timestamp to be
                 refresh per access. default: False
+            * **thread_safe (bool)**: tell decorator to use thread safe lock. default=False
+            * **max_mem_size (int)**: max mem size in bytes, all cache values in cache structure. default=None which means no limit
         """
-        self.cache = cache_class(maxsize, ttl, refresh_ttl=refresh_ttl)
+        self.cache = cache_class(
+            maxsize,
+            ttl,
+            refresh_ttl=refresh_ttl,
+            thread_safe=thread_safe,
+            max_mem_size=max_mem_size,
+        )
         self.maxsize = maxsize
         self.skip_args = skip_args
 
